#!/bin/bash

# read sql settings from configuration file
if [ -f ./conf/mcore-sql-schema.conf ]
then
	source ./conf/mcore-sql-schema.conf
fi

# include everything from include
for inc in $(find ./include -type f -name \*.sh)
do
	source "${inc}"
done

die()
{
	echo "ERROR: $@"
	exit 1
}

mysql_command()
{
	local sql_user
	local sql_pass
	local sql_host
	local sql_db
	local sql_command

	sql_user=$1
	sql_pass=$2
	sql_host=$3
	sql_db=$4
	sql_command=$5

	#fallback (sql_opts not, they are optional )
	if [ -z "${sql_user}" \
		-o -z "${sql_pass}" \
		-o -z "${sql_host}" \
		-o -z "${sql_db}" \
		-o -z "${sql_command}" ]
	then
		echo
		echo "Usage:"
		echo -e "      mysql_command user pass host db \"sql-command\""
		echo
		return 1
	fi

	mysql \
		--user="${sql_user}" \
		--password="${sql_pass}" \
		--host="${sql_host}" \
		--database="${sql_db}" \
		--batch \
		--skip-column-names \
		--execute="${sql_command}" \
		|| return 1

	return 0
}

mysqldo()
{
	mysql_command \
		"${MCORE_SQL_USER}" \
		"${MCORE_SQL_PASS}" \
		"${MCORE_SQL_HOST}" \
		"${MCORE_SQL_DB}" \
		"$@" \
		&& return 0 || return 1
}

# creates a mysql table with help of a defined array TABLE_DEFINITION:
# array values are a struct like this:
# first array element is the table name, everything else are column descriptions
# TABLE_DEFINITION=(
#		"table_name"
#			"column1_name column1_argvs"
#			"column2_name column2_argvs"
#			"columnN_name columnN_argvs"
#	)
sql_create_table()
{
	local table
	local count
	local i
	local line
	local opts
	local column
	local autoincrement
	local primary
	local create

	table="${TABLE_DEFINITION[0]}"
	count="${#TABLE_DEFINITION[*]}"

	for ((i=1; i<count; i++))
	do
		line=( ${TABLE_DEFINITION[${i}]} )
		column="${line[0]}"
		opts="${line[*]:1}"

		case ${opts} in
			*"PRIMARY KEY"*) opts="${opts//PRIMARY KEY}"; primary="${column}" ;;
			*"primary key"*) opts="${opts//primary key}"; primary="${column}" ;;
		esac

		case ${opts} in
			*AUTO_INCREMENT*) opts="${opts//AUTO_INCREMENT}"; autoincrement="${column} ${opts}" ;;
			*auto_increment*) opts="${opts//auto_increment}"; autoincrement="${column} ${opts}" ;;
		esac

		if [[ -z ${create} ]]
		then
			create="${column} ${opts}"
		else
			create+=", ${column} ${opts}"
		fi
	done
	mysqldo "CREATE TABLE ${MCORE_SQL_DB}.${table} ( ${create} );" || die "creating table '${table}'"

	if [[ -n ${primary} ]]
	then
		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} ADD PRIMARY KEY (${primary});" || die "adding primary key to table '${table}'"
	fi
	if [[ -n ${autoincrement} ]]
	then
		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} MODIFY ${autoincrement} AUTO_INCREMENT;" || die "adding auto_increment to table '${table}'"
	fi
}

# TABLE_DEFINITION=(
#		"table_name"
#			"column1_name column1_argvs"
#			"column2_name column2_argvs"
#			"columnN_name columnN_argvs"
#
# eg.    "description varchar(255) DEFAULT NULL AFTER value"
#	)
#
sql_add_column()
{
	local table
	local count
	local i
	local line
	local opts
	local column
	local autoincrement
	local primary
	local current_primary
	local current_autoincrement
	local current_opts

	table="${TABLE_DEFINITION[0]}"
	count="${#TABLE_DEFINITION[*]}"

	for ((i=1; i<count; i++))
	do
		line=( ${TABLE_DEFINITION[${i}]} )
		column="${line[0]}"
		opts="${line[*]:1}"

		case ${opts} in
			*"PRIMARY KEY"*) opts="${opts//PRIMARY KEY}"; primary="${column}" ;;
			*"primary key"*) opts="${opts//primary key}"; primary="${column}" ;;
		esac

		case ${opts} in
			*AUTO_INCREMENT*) opts="${opts//AUTO_INCREMENT}"; autoincrement="${column} ${opts}" ;;
			*auto_increment*) opts="${opts//auto_increment}"; autoincrement="${column} ${opts}" ;;
		esac

		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} ADD ${column} ${opts};" || die "add column '${table}'"
	done

	# delete an existing auto_increment first
	if [[ -n ${autoincrement} ]]
	then
		i=( $(mysqldo 'SHOW COLUMNS FROM '${MCORE_SQL_DB}.${table}' WHERE `Extra`="auto_increment"') )
		current_autoincrement="${i[0]}"
		if [[ ${current_autoincrement} != ${column} ]] && [[ -n ${current_autoincrement} ]]
		then
			current_opts=( "${i[1]}" ) # type eg int(11)
			# null or not
			case "${i[2]}" in
				NO|no) current_opts+=( "NOT NULL" ) ;;
				YES|yes) current_opts+=( "DEFAULT ${i[4]}" ) ;; # should never match, but to be save
			esac
			mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} CHANGE ${current_autoincrement} ${current_autoincrement} ${current_opts};" || die "removing auto_increment to table '${table}'"
		fi
	fi

	if [[ -n ${primary} ]]
	then
		# only run this if the primary key is not already the same column
		i=( $(mysqldo 'SHOW COLUMNS FROM '${MCORE_SQL_DB}.${table}' WHERE `Key`="PRI"') )
		current_primary="${i[0]}"
		if [[ -n ${current_primary} ]]
		then
			mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} DROP PRIMARY KEY;" || die "adding primary key to table '${table}'"
		fi
		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} ADD PRIMARY KEY (${primary});" || die "adding primary key to table '${table}'"
	fi
	if [[ -n ${autoincrement} ]]
	then
		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} MODIFY ${autoincrement} AUTO_INCREMENT;" || die "adding auto_increment to table '${table}'"
	fi
}

# TABLE_DEFINITION=(
#		"table_name"
#			"column1_name column1_argvs"
#			"column2_name column2_argvs"
#			"columnN_name columnN_argvs"
#	)
sql_modify_column()
{
	local table
	local count
	local i
	local line
	local opts
	local column
	local autoincrement
	local primary
	local current_primary
	local current_autoincrement
	local current_opts

	table="${TABLE_DEFINITION[0]}"
	count="${#TABLE_DEFINITION[*]}"

	for ((i=1; i<count; i++))
	do
		line=( ${TABLE_DEFINITION[${i}]} )
		column="${line[0]}"
		opts="${line[*]:1}"

		case ${opts} in
			*"PRIMARY KEY"*) opts="${opts//PRIMARY KEY}"; primary="${column}" ;;
			*"primary key"*) opts="${opts//primary key}"; primary="${column}" ;;
		esac

		case ${opts} in
			*AUTO_INCREMENT*) opts="${opts//AUTO_INCREMENT}"; autoincrement="${column} ${opts}" ;;
			*auto_increment*) opts="${opts//auto_increment}"; autoincrement="${column} ${opts}" ;;
		esac

		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} MODIFY ${column} ${opts};" || die "modify table '${table}'"
	done

	# delete an existing auto_increment first
	if [[ -n ${autoincrement} ]]
	then
		i=( $(mysqldo 'SHOW COLUMNS FROM '${MCORE_SQL_DB}.${table}' WHERE `Extra`="auto_increment"') )
		current_autoincrement="${i[0]}"
		if [[ ${current_autoincrement} != ${column} ]] && [[ -n ${current_autoincrement} ]]
		then
			current_opts=( "${i[1]}" ) # type eg int(11)
			# null or not
			case "${i[2]}" in
				NO|no) current_opts+=( "NOT NULL" ) ;;
				YES|yes) current_opts+=( "DEFAULT ${i[4]}" ) ;; # should never match, but to be save
			esac
			mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} CHANGE ${current_autoincrement} ${current_autoincrement} ${current_opts};" || die "removing auto_increment to table '${table}'"
		fi
	fi

	if [[ -n ${primary} ]]
	then
		# only run this if the primary key is not already the same column
		i=( $(mysqldo 'SHOW COLUMNS FROM '${MCORE_SQL_DB}.${table}' WHERE `Key`="PRI"') )
		current_primary="${i[0]}"
		if [[ -n ${current_primary} ]]
		then
			mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} DROP PRIMARY KEY;" || die "adding primary key to table '${table}'"
		fi
		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} ADD PRIMARY KEY (${primary});" || die "adding primary key to table '${table}'"
	fi
	if [[ -n ${autoincrement} ]]
	then
		mysqldo "ALTER TABLE ${MCORE_SQL_DB}.${table} MODIFY ${autoincrement} AUTO_INCREMENT;" || die "adding auto_increment to table '${table}'"
	fi
}

# VALUE_DEFINITION=(
#		"table_name, value_column_name, bool_column_name, id_column_name"
#			"value1, bool1"
#			"value2, bool2"
#			"valueN, boolN"
#	)
sql_set_value()
{
	local line
	local meta
	local table
	local column_value
	local column_bool
	local column_id
	local value
	local bool
	local count
	local i
	local id

	count="${#VALUE_DEFINITION[*]}"
	line="${VALUE_DEFINITION[0]}"
	for i in ${line//,/ }
	do
		meta+=( "${i}" )
	done

	table="${meta[0]}"
	column_value="${meta[1]}"
	column_bool="${meta[2]}"
	column_id="${meta[3]}"

	for ((i=1; i<count; i++))
	do
		line="${VALUE_DEFINITION[${i}]}"
		value="${line%%,*}"
		bool="${line##*,}"
		# remove any leading whitespaces
		bool="${bool// }"

		# now check if the value already exist in the database
		id=$(mysqldo "SELECT ${column_id} FROM ${MCORE_SQL_DB}.${table} WHERE ${column_value}='${value}';")
		if [[ -n ${id} ]]
		then
			# update value in database
			mysqldo "UPDATE ${MCORE_SQL_DB}.${table} SET ${column_value}='${value}', ${column_bool}='${bool}' WHERE ${column_id}='${id}';" || die "update value in table ${table}"
		else
			# insert into the database
			mysqldo "INSERT INTO ${MCORE_SQL_DB}.${table} (${column_value}, ${column_bool}) VALUES ( '${value}', '${bool}' );" || die "insert value into table ${table}"
		fi
	done
}

create_database()
{
	echo "Creating database schema version ${MCORE_SQL_DB_SCHEMA_VERSION}"
	create_database_schema

	# update database schema version
	mysqldo "INSERT INTO ${MCORE_SQL_DB}.meta_schema(version) VALUES ('${MCORE_SQL_DB_SCHEMA_VERSION}') ON DUPLICATE KEY UPDATE version='${MCORE_SQL_DB_SCHEMA_VERSION}';" || die "updating schema version"
}

update_database()
{
	local current_sql_db_schema_version
	local i

	current_sql_db_schema_version=$(mysqldo "SELECT MAX(version) from meta_schema;")

	if [[ ${current_sql_db_schema_version} = ${MCORE_SQL_DB_SCHEMA_VERSION} ]]
	then
		echo "Database schema is already up to date"
		return 0
	fi

	# run every update_to_database_version_X script which exist up to version MCORE_SQL_DB_SCHEMA_VERSION
	for i in $(seq $((${current_sql_db_schema_version}+1)) ${MCORE_SQL_DB_SCHEMA_VERSION})
	do
		if [[ -n $(typeset -f update_to_database_schema_version_${i}) ]]
		then
			echo "Running update_to_database_schema_version_${i}()"
			update_to_database_schema_version_${i} || die
		fi
	done

	# update database schema version
	mysqldo "INSERT INTO ${MCORE_SQL_DB}.meta_schema(version) VALUES ('${MCORE_SQL_DB_SCHEMA_VERSION}') ON DUPLICATE KEY UPDATE version='${MCORE_SQL_DB_SCHEMA_VERSION}';" || die "updating schema version"
}

fill_database()
{
	local current_sql_db_values_version
	local i

	current_sql_db_values_version=$(mysqldo "SELECT MAX(version) from meta_values;")

	if [[ ${current_sql_db_values_version} = ${MCORE_SQL_DB_VALUES_VERSION} ]]
	then
		echo "Database values are already up to date"
		return 0
	fi

	echo "Updating database default values to version ${MCORE_SQL_DB_VALUES_VERSION}"
	create_database_values
	# update database values version
	mysqldo "INSERT INTO ${MCORE_SQL_DB}.meta_values(version) VALUES ('${MCORE_SQL_DB_VALUES_VERSION}') ON DUPLICATE KEY UPDATE version='${MCORE_SQL_DB_VALUES_VERSION}';" || die "updating values version"
}

usage()
{
		echo "Usage: $0 [command]"
		echo "Commands:"
		echo "   create - creates a new mcore database, the database 'mcore' must exist"
		echo "   update - updates the current database to the newest patchlevel"
}

case $1 in
	create)
		create_database
		fill_database
		;;

	update)
		update_database
		fill_database
		;;

	*) usage ;;
esac
